import argparse
import asyncio
import logging
import struct
from typing import Any, Optional, Tuple
from urllib.parse import urlparse

from aiokdb import cv, logger
from aiokdb.server import (
    BaseContext,
    CredentialsException,
    KdbReader,
    KdbWriter,
    reader_to_context_task,
)


class ClientContext(BaseContext):
    pass


background_tasks = set()


# KDB client code
async def open_qipc_connection(
    host: str = "127.0.0.1",
    port: int = 8890,
    user: Optional[str] = None,
    password: Optional[str] = None,
    uri: Optional[str] = None,
    context: Optional[ClientContext] = None,
    ver: int = 3,
) -> Tuple[KdbReader, KdbWriter]:
    if uri:  #  uri takes precedence if provided
        pr = urlparse(uri)
        if pr.hostname:
            host = pr.hostname
        if pr.port:
            port = pr.port
        user = pr.username
        password = pr.password

    reader, writer = await asyncio.open_connection(host, port)

    auth = ""
    if user:
        auth = auth + user
    if password:
        auth = auth + ":" + password

    writer.write(auth.encode() + struct.pack("<B", ver) + b"\000")
    await writer.drain()

    try:
        data = await reader.readexactly(1)  # negotiated version, if auth correct
        remote_ver = struct.unpack("<B", data)[0]
        if remote_ver != ver:
            raise Exception(f"expected version {ver}, server gave {remote_ver}")
        logger.debug(f"Connected OK, remote_ver={remote_ver}")
    except asyncio.IncompleteReadError as e:
        raise CredentialsException(e)

    q_reader = KdbReader(reader)
    q_writer = KdbWriter(writer, q_reader, version=remote_ver, context=context)
    if context is not None:
        task = asyncio.create_task(reader_to_context_task(q_writer, q_reader, context))
        background_tasks.add(task)
        task.add_done_callback(background_tasks.discard)

    return q_reader, q_writer


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser()
    parser.add_argument("--host", default="localhost")
    parser.add_argument("--port", default=8890, type=int)
    parser.add_argument("--user", default="user")
    parser.add_argument("--password")
    args = parser.parse_args()

    async def main(args: Any) -> None:
        r, w = await open_qipc_connection(
            host=args.host, port=args.port, user=args.user, password=args.password
        )
        while True:
            obj = await w.sync_req(cv(input("Enter query: ")))
            print(f"Got object {obj}")
        return None

    asyncio.run(main(args))



import uuid

from aiokdb import TypeEnum, cv, d9, ka, kj, kk, ks, ktn, xd, xt
from aiokdb.extras import ktni, ktns
from aiokdb.format import AsciiFormatter


def test_format_unkeyed_table() -> None:
    fmt = AsciiFormatter(height=8)

    # q)-8!([]a:enlist 2i;b:enlist 3i)
    ks = ktn(TypeEnum.KS).appendS("a", "b")
    kv = ktn(TypeEnum.K)
    kv.kK().extend([ktn(TypeEnum.KI, 1), ktn(TypeEnum.KI, 1)])
    kv.kK()[0].kI()[0] = 2
    kv.kK()[1].kI()[0] = 3
    t = xt(xd(ks, kv))

    assert fmt.format(t) == "a b\n---\n2 3"

    # first col pushed out by col name, 2nd by value width
    ks = ktn(TypeEnum.KS).appendS("alpha", "b", "cheese")
    kv = ktn(TypeEnum.K)
    kv.kK().extend([ktn(TypeEnum.KI, 3), ktn(TypeEnum.KI, 3), ktn(TypeEnum.KI, 3)])
    kv.kK()[0].kI()[1] = 1
    kv.kK()[0].kI()[2] = 2
    kv.kK()[1].kI()[1] = 17
    t = xt(xd(ks, kv))

    expected = "alpha b  cheese\n---------------\n0     0  0     \n1     17 0     \n2     0  0     "
    assert fmt.format(t) == expected

    # introducing the ... fold at halfway
    ks = ktn(TypeEnum.KS).appendS("long")
    kv = ktn(TypeEnum.K)
    kv.kK().extend([ktn(TypeEnum.KI, 50)])
    kv.kK()[0].kI()[49] = 49
    t = xt(xd(ks, kv))
    expected = "long\n----\n0   \n0   \n... \n0   \n49  "
    assert fmt.format(t) == expected


def test_format_keyed_table() -> None:
    expected = "a| b\n-|--\n2| 3"

    t = d9(
        bytes.fromhex(
            "010000003f000000636200630b00010000006100000001000000060001000000020000006200630b0001000000620000000100000006000100000003000000"
        )
    )
    fmt = AsciiFormatter()
    assert fmt.format(t) == expected


def test_format_keyed_table_nested_str() -> None:
    expected = "envelope_id                         | payload time                         \n------------------------------------|--------------------------------------\n2d948578-e9d6-79a2-8207-9df7a71f0b3b| abc     2024.05.14D23:13:19:044908000\n409031f3-b19c-6770-ee84-6e9369c98697| xy      2024.05.14D23:13:19:044908000"
    # keyed table
    # dictionary with tables for keys and values
    key_hdr = ktns(TypeEnum.KS, "envelope_id")
    key_val = kk(ktn(TypeEnum.UU))

    val_hdr = ktns(TypeEnum.KS, "payload", "time")
    val_val = kk(ktn(TypeEnum.K), ktn(TypeEnum.KP))
    kt = xd(xt(xd(key_hdr, key_val)), xt(xd(val_hdr, val_val)))

    kt.kkey()["envelope_id"].kU().extend(
        [
            uuid.UUID("2d948578-e9d6-79a2-8207-9df7a71f0b3b"),
            uuid.UUID("409031f3-b19c-6770-ee84-6e9369c98697"),
        ]
    )
    kt.kvalue()["payload"].kK().extend([cv("abc"), cv("xy")])
    kt.kvalue()["time"].kJ().extend([769043599044908000, 769043599044908000])

    assert AsciiFormatter().format(kt) == expected


def test_format_dict() -> None:
    d = xd(ktni(TypeEnum.KJ, 3, 612, 6), ktns(TypeEnum.KS, "hi", "p", "dog"))
    fmt = AsciiFormatter()
    assert fmt.format(d) == "3  | hi\n612| p\n6  | dog"

    d = xd(ktni(TypeEnum.KJ, 3, 612, 6), kk(kj(56), ks("xray"), ka(-TypeEnum.UU)))
    assert (
        fmt.format(d) == "3  | 56\n612| xray\n6  | 00000000-0000-0000-0000-000000000000"
    )


def test_format_atoms() -> None:
    fmt = AsciiFormatter(height=8)

    p = ka(-TypeEnum.KP)
    p.j(1000000000)
    assert fmt.format(p) == "2000.01.01D00:00:01:000000000"

    # square brackets to show the mixed-type nature of K-arrays
    p = kk(kj(5), ks("hello"))
    assert fmt.format(p) == "[5, hello]"

    p = ktni(TypeEnum.KJ, *range(20))
    assert fmt.format(p) == "0 1 ... 18 19"



import inspect
import re
import uuid
from typing import cast

from aiokdb import KIntArray, KLongArray, KObj, TypeEnum, kb, kj, kk, kNil, ks, ktn, tn
from aiokdb.client import ClientContext
from aiokdb.server import KdbWriter, ServerContext


# these are python friendly constructors for vector types, incomplete
# not sure if there is a better way
def ktni(t: TypeEnum, *ints: int) -> KObj:
    v = ktn(t)
    if t == TypeEnum.KG:
        v.kG().extend(ints)
    elif t == TypeEnum.KH:
        v.kH().extend(ints)
    elif isinstance(v, KIntArray):
        v.kI().extend(ints)
    elif isinstance(v, KLongArray):
        v.kJ().extend(ints)
    else:
        raise ValueError(f"No int array initialiser for {tn(t)}")
    return v


def ktns(t: TypeEnum, *ss: str) -> KObj:
    v = ktn(t)
    if t == TypeEnum.KS:
        for s in ss:
            v.appendS(s)
    else:
        raise ValueError(f"No str array initialiser for {tn(t)}")
    return v


def ktnu(*uuids: uuid.UUID) -> KObj:
    v = ktn(TypeEnum.UU)
    v.kU().extend(uuids)
    return v


def ktnb(*bools: bool) -> KObj:
    v = ktn(TypeEnum.KB)
    v.kB().extend(bools)
    return v


# We assume commands of this pattern "func[arg1;arg2;argN]"
# which we convert to a k-array with symbol as the first argument, and remaining
# arguments (badly) converted to k-atoms. KDBs parser is a work of art, and this is
# admittedly a poor immitation.
def _string_to_functional(cmd: KObj) -> KObj:
    if not isinstance(cmd, KObj):
        raise ValueError("Expected KObj")
    if cmd.t != TypeEnum.KC:
        raise ValueError("Expected char vector input")
    c = cmd.aS()
    result = re.match(r"(.+)\[(.*)\]", c)
    if not result:
        raise ValueError("error parsing: {}".format(cmd))

    functional = [ks(result.group(1))]
    if result.group(2):
        cmd_args = result.group(2).split(";")
        print(cmd_args)
        for arg in cmd_args:
            if len(arg) > 1 and arg[0] == "`":
                # symbol
                functional.append(ks(arg[1:]))
            elif len(arg) == 2 and arg[1] == "b":
                # bool atom
                functional.append(kb(False if arg[0] == "0" else True))
            elif arg.isnumeric():
                functional.append(kj(int(arg)))
            else:
                raise ValueError(f"error parsing arg: {arg} in command {c}")
    else:
        functional.append(kNil)
    return kk(*functional)


# this tries to offer basic eval support for function dispatch within a server
# users expect the input to the server to hit eval, which will parse the arguments to
# kdb types, then dispatch to a function.
class MagicContext:
    async def on_sync_request(self, cmd: KObj, dotzw: KdbWriter) -> KObj:  # .z.pg
        # kdb clients usually present RPC to server as a string, evaluated
        # with value, although it is possible to have arbitary objects here
        if cmd.t == TypeEnum.KC:
            cmd = _string_to_functional(cmd)

        # cmd should now be a k, with symbol as first argument
        fn = cmd.kK()[0].aS()
        # map kx function name (containing periods) to python and do getattr
        # this is behind the login check, so I'm not concerned about RPC
        args = kk(*cmd.kK()[1:])

        fnpy = fn.lstrip(".").replace(".", "__")  # drop initial dot, snake dots
        try:
            f = getattr(self, fnpy)
            k = f(args, dotzw)
            if inspect.isawaitable(k):
                return cast(KObj, await k)
            return cast(KObj, k)
        except AttributeError:
            raise ValueError(f"No python function {fnpy} found from {fn}")


class MagicServerContext(MagicContext, ServerContext):
    pass


class MagicClientContext(MagicContext, ClientContext):
    pass



import asyncio
from typing import List, Optional

import pytest

from aiokdb import KException, KObj, MessageType, TypeEnum, cv, kj, kNil
from aiokdb.client import open_qipc_connection
from aiokdb.extras import MagicClientContext, MagicServerContext, _string_to_functional
from aiokdb.server import CredentialsException, KdbWriter, ServerContext, start_qserver


@pytest.mark.asyncio
async def test_client_noauth_defaultport() -> None:
    context = ServerContext()
    server = await start_qserver(8890, context)

    client_rd, client_wr = await open_qipc_connection()

    with pytest.raises(KException, match="nyi handling"):
        await client_wr.sync_req(cv("1+2"))

    client_wr.close()
    await client_wr.wait_closed()

    # check supplied auth ignored when noauth configured
    client_rd, client_wr = await open_qipc_connection(user="troy", password="tango")
    client_wr.close()
    await client_wr.wait_closed()

    server.close()
    await server.wait_closed()


# Please manually verify auth changes against KDB.
#
# Python server, KDB client
#   $ python -m qaio.server
#   $ ~/kdb/q
#   q)hopen ":localhost:8890"           INFO:root:q-0 process_login ver=6 user=chrisshucksmith password=None
#   q)hopen ":localhost:8890:blah"      INFO:root:q-1 process_login ver=6 user=blah password=None
#   q)hopen ":localhost:8890:blah:boo"  INFO:root:q-2 process_login ver=6 user=blah password=***
#
# Python client, KDB server
#   $ ~/kdb/q -p 8890
#   q).z.pw:{[user;pswd]1 "user: ",(string user)," password: ",pswd,"\n";1b}
#   $ python -m qaio.client                          q) user: user password:
#   $ python -m qaio.client --user u1                q) user: u1 password:
#   $ python -m qaio.client --user u1 --password p2  q) user: u1 password: p2


@pytest.mark.asyncio
async def test_client_auth_defined_port() -> None:
    # echo back client request
    class EchoServerContext(ServerContext):
        async def on_sync_request(self, cmd: KObj, dotzw: KdbWriter) -> KObj:
            return cmd

    context = EchoServerContext("tangoxray")
    server = await start_qserver(6778, context)

    # check correct auth succeeds
    client_rd, client_wr = await open_qipc_connection(
        port=6778, user="troy", password="tangoxray"
    )
    kr = await client_wr.sync_req(cv("1+2"))
    assert kr.aS() == "1+2"

    client_wr.close()
    await client_wr.wait_closed()

    # check wrong auth fails
    with pytest.raises(CredentialsException):
        client_rd, client_wr = await open_qipc_connection(
            port=6778, user="troy", password="xyz"
        )

    # hopen can be sent without a password
    with pytest.raises(CredentialsException):
        client_rd, client_wr = await open_qipc_connection(
            port=6778, user="xyz", password=None
        )

    # hopen can be sent with empty username (encodes as "")
    with pytest.raises(CredentialsException):
        client_rd, client_wr = await open_qipc_connection(
            port=6778, user=None, password=None
        )

    server.close()
    await server.wait_closed()


@pytest.mark.asyncio
async def test_client_auth_uri() -> None:
    context = ServerContext("tango")
    server = await start_qserver(6779, context)

    # can connect by url
    client_rd, client_wr = await open_qipc_connection(
        uri="kdb://qq:tango@localhost:6779"
    )

    client_wr.close()
    await client_wr.wait_closed()

    server.close()
    await server.wait_closed()


@pytest.mark.asyncio
async def test_server_async() -> None:
    class AsyncServerContext(ServerContext):
        async def on_sync_request(self, cmd: KObj, dotzw: KdbWriter) -> KObj:
            # async message sent during sync request handling
            dotzw.write(kj(1), MessageType.ASYNC)
            # return value becomes the sync response
            return kj(2)

    context = AsyncServerContext("tangoxray")
    server = await start_qserver(6778, context)

    client_rd, client_wr = await open_qipc_connection(
        port=6778, user="troy", password="tangoxray"
    )

    ooob: List[KObj] = []
    kr = await client_wr.sync_req(cv("1+2"), ooob=ooob.append)
    assert kr.aJ() == 2

    assert len(ooob) == 1
    assert ooob[0].aJ() == 1

    client_wr.close()
    await client_wr.wait_closed()

    server.close()
    await server.wait_closed()


def test_extras_parse_commands() -> None:
    with pytest.raises(ValueError):
        _string_to_functional(kj(4))

    with pytest.raises(ValueError):
        _string_to_functional("bare string not KObj")  # type: ignore

    c = cv("blah[3;45;`silly;0b]")
    k = _string_to_functional(c)

    assert len(k) == 5
    assert k.t == TypeEnum.K
    bits = k.kK()
    assert bits[0].aS() == "blah"
    assert bits[1].aJ() == 3
    assert bits[2].aJ() == 45
    assert bits[3].aS() == "silly"
    assert bits[4].aB() is False

    # KDB doesn't have no argument functions, they are passed [::]
    k = _string_to_functional(cv("regular.function[]"))
    assert len(k) == 2
    assert k.t == TypeEnum.K
    bits = k.kK()
    assert bits[0].aS() == "regular.function"
    assert bits[1] is kNil


@pytest.mark.asyncio
async def test_extras_MagicServerContext() -> None:
    # defining the function within a MagicServerContext is enough to make it callable
    class MyContext(MagicServerContext):
        async def mynamespace__myfunction(self, args: KObj, dotzw: KdbWriter) -> KObj:
            return kj(args.kK()[0].aJ())

        def regular__function(self, args: KObj, dotzw: KdbWriter) -> KObj:
            return kj(34)

    server = await start_qserver(6778, MyContext())
    client_rd, client_wr = await open_qipc_connection(port=6778)

    kr = await client_wr.sync_req(cv(".mynamespace.myfunction[45]"))
    assert kr.aJ() == 45

    kr = await client_wr.sync_req(cv("regular.function[]"))
    assert kr.aJ() == 34

    with pytest.raises(
        KException,
        match="No python function mynamespace__notafunction found from .mynamespace.notafunction",
    ):
        await client_wr.sync_req(cv(".mynamespace.notafunction[45]"))

    assert (await client_wr.sync_req(cv("regular.function[]"))).aJ() == 34

    client_wr.close()
    await client_wr.wait_closed()

    server.close()
    await server.wait_closed()


@pytest.mark.asyncio
async def test_server_calls_client() -> None:
    class TestServerContext(MagicServerContext):
        stored_handle: Optional[KdbWriter] = None

        async def storehandle(self, args: KObj, dotzw: KdbWriter) -> KObj:
            self.stored_handle = dotzw
            return kNil

        async def checkafter(self, args: KObj, dotzw: KdbWriter) -> KObj:
            return kj(32)

    server = await start_qserver(6778, tsc := TestServerContext())

    class TestClientContext(MagicClientContext):
        async def dosums(self, args: KObj, dotzw: KdbWriter) -> KObj:
            return kj(42)

    client_rd, client_wr = await open_qipc_connection(
        port=6778, user="troy", password="tangoxray", context=TestClientContext()
    )

    assert tsc.stored_handle is None
    kr = await client_wr.sync_req(cv("storehandle[]"))
    assert kr == kNil
    assert tsc.stored_handle is not None

    # call client method from the server
    assert (await tsc.stored_handle.sync_req(cv("dosums[]"))).aJ() == 42

    assert (await client_wr.sync_req(cv("checkafter[]"))).aJ() == 32

    client_wr.close()
    await client_wr.wait_closed()

    server.close()
    await server.wait_closed()


@pytest.mark.asyncio
async def test_client_speaks_first() -> None:
    class BannerServerContext(MagicServerContext):
        async def banner(self, args: KObj, dotzw: KdbWriter) -> KObj:
            return kj(32)

    server = await start_qserver(6778, BannerServerContext())

    fut = asyncio.get_running_loop().create_future()

    class TestClientContext(MagicClientContext):
        async def writer_available(self, dotzw: KdbWriter) -> None:
            self.writer = dotzw
            print("connected - sending subscribe")
            try:
                r = await dotzw.sync_req(cv("banner[]"))
                print(f"recieved banner {r.aJ()}")
                fut.set_result(r)
            except Exception:
                print("banner login rejected")

    client_rd, client_wr = await open_qipc_connection(
        port=6778, user="troy", password="tangoxray", context=TestClientContext()
    )

    r = await fut
    assert r.aJ() == 32

    client_wr.close()
    await client_wr.wait_closed()

    server.close()
    await server.wait_closed()



import array
import enum
import logging
import struct
import uuid
from collections.abc import MutableSequence, Sequence
from typing import Any, Dict, List, Tuple, Type, Union, cast

from aiokdb.adapter import BoolByteAdaptor
from aiokdb.compress import decompress

__all__ = [
    "b9",
    "d9",
    "ktn",
    "kk",
    "cv",
    "MessageType",
    "Nulls",
    "Infs",
    "TypeEnum",
    "KObj",
    "xd",
    "xt",
    "ka",
    "kb",
    "kc",
    "kg",
    "kh",
    "ki",
    "kj",
    "ks",
    "kuu",
]

# mypy: disallow-untyped-defs

logger = logging.getLogger(__name__)


class WrongTypeForOperationError(TypeError):
    pass


class AttrEnum(enum.IntEnum):
    NONE = 0
    SORTED = 1
    UNIQUE = 2
    PARTED = 3
    GROUPED = 4


class MessageType(enum.IntEnum):
    ASYNC = 0
    SYNC = 1
    RESPONSE = 2


class Nulls:
    h: int = -32768
    i: int = -2147483648
    j: int = -9223372036854775808
    c: str = " "
    e: float = float("nan")
    f: float = float("nan")


class Infs:
    h: int = 32767
    i: int = 2147483647
    j: int = 9223372036854775807
    e: float = float("inf")
    f: float = float("inf")


class TypeEnum(enum.IntEnum):
    #    type bytes qtype     ctype  accessor
    K = 0  #   *    K         K
    KB = 1  #  1    boolean   char   kG
    UU = 2  #  16   guid      U      kU
    KG = 4  #  1    byte      char   kG
    KH = 5  #  2    short     short  kH
    KI = 6  #  4    int       int    kI
    KJ = 7  #  8    long      long   kJ
    KE = 8  #  4    real      float  kE
    KF = 9  #  8    float     double kF
    KC = 10  # 1    char      char   kC
    KS = 11  # *    symbol    char*  kS
    KP = 12  # 8    timestamp long   kJ (nanoseconds from 2000.01.01)
    KM = 13  # 4    month     int    kI (months from 2000.01.01)
    KD = 14  # 4    date      int    kI (days from 2000.01.01)
    KZ = 15  # 8    datetime  double kF (DO NOT USE)
    KN = 16  # 8    timespan  long   kJ (nanoseconds)
    KU = 17  # 4    minute    int    kI
    KV = 18  # 4    second    int    kI
    KT = 19  # 4    time      int    kI (millisecond)
    XT = 98  # table
    XD = 99  # dict
    FN = 100  # function
    NIL = 101  # nil item
    SD = 127  # sorted dict
    KRR = -128  # error


# Remember to special case TypeEnum.KS and TypeEnum.K
ATOM_LENGTH: Dict[int, int] = {
    TypeEnum.KB: 1,
    TypeEnum.UU: 16,
    TypeEnum.KG: 1,
    TypeEnum.KH: 2,
    TypeEnum.KI: 4,
    TypeEnum.KJ: 8,
    TypeEnum.KE: 4,
    TypeEnum.KF: 8,
    TypeEnum.KC: 1,
    TypeEnum.KP: 8,
    TypeEnum.KM: 4,
    TypeEnum.KD: 4,
    TypeEnum.KN: 8,
    TypeEnum.KU: 4,
    TypeEnum.KV: 4,
    TypeEnum.KT: 4,
    TypeEnum.KZ: 8,
    -TypeEnum.NIL: 1,  # TODO: unhack this
}


class KContext:
    def __init__(self) -> None:
        self.symbols: Dict[str, int] = {}
        self.symbols_enc: Dict[int, Tuple[str, bytes]] = {}

    def ss(self, s: str) -> int:
        # we don't want any surprises trying to serialise non-ascii symbols later
        # so force non-ascii characters out now
        bs = bytes(s, "ascii") + b"\x00"
        idx = self.symbols.setdefault(s, len(self.symbols))
        # TODO this should be a list
        self.symbols_enc[idx] = (s, bs)
        return idx


DEFAULT_CONTEXT = KContext()


def tn(t: int) -> str:
    te = t
    if te != TypeEnum.KRR:
        te = abs(t)
    if te >= 20 and te < 40:
        return "Enum ({t})"
    return f"{TypeEnum(te).name} ({t})"


class KObj:
    def __init__(
        self,
        t: int = 0,
        context: KContext = DEFAULT_CONTEXT,
        sz: int = 0,
        attr: int = 0,
    ) -> None:
        self.t = t
        self.attrib = attr
        self.context: KContext = context

    def _paysz(self) -> int:
        return 1

    def _databytes(self) -> bytes:
        return struct.pack("<b", self.t)

    def _tn(self) -> str:
        return tn(self.t)

    def _te(self) -> Exception:
        return WrongTypeForOperationError(f"Not available for {self._tn()}")

    def __len__(self) -> int:
        raise self._te()

    # atom setters
    def ss(self, s: str) -> "KObj":
        raise self._te()

    def b(self, b: bool) -> "KObj":
        raise self._te()

    def c(self, c: str) -> "KObj":
        raise self._te()

    def g(self, g: int) -> "KObj":
        raise self._te()

    def h(self, h: int) -> "KObj":
        raise self._te()

    def i(self, i: int) -> "KObj":
        raise self._te()

    def j(self, j: int) -> "KObj":
        raise self._te()

    def uu(self, uu: uuid.UUID) -> "KObj":
        raise self._te()

    # atom getters
    def aB(self) -> bool:
        raise self._te()

    def aG(self) -> int:
        raise self._te()

    def aI(self) -> int:
        raise self._te()

    def aH(self) -> int:
        raise self._te()

    def aJ(self) -> int:
        raise self._te()

    def aS(self) -> str:
        raise self._te()

    def aU(self) -> uuid.UUID:
        raise self._te()

    def aE(self) -> float:
        raise self._te()

    def aF(self) -> float:
        raise self._te()

    def aC(self) -> str:
        raise self._te()

    # vector getters
    def kK(self) -> "MutableSequence[KObj]":
        raise self._te()

    def kU(self) -> "MutableSequence[uuid.UUID]":
        raise self._te()

    def kB(self) -> "MutableSequence[bool]":
        raise self._te()

    def kG(self) -> "MutableSequence[int]":
        raise self._te()

    def kH(self) -> "MutableSequence[int]":
        raise self._te()

    def kI(self) -> "MutableSequence[int]":
        raise self._te()

    def kJ(self) -> "MutableSequence[int]":
        raise self._te()

    def kE(self) -> "MutableSequence[float]":
        raise self._te()

    def kF(self) -> "MutableSequence[float]":
        raise self._te()

    def kC(self) -> array.array:  # type: ignore[type-arg]
        raise self._te()

    def kS(self) -> "Sequence[str]":
        raise self._te()

    # dictionary/flip
    def kkey(self) -> "KObj":
        raise self._te()

    def kvalue(self) -> "KObj":
        raise self._te()

    def __getitem__(self, item: Union[int, str]) -> "KObj":
        raise self._te()

    # TODO: clean this up by having kS() return a MutableSequence(str) that
    # writes through to the int array
    def appendS(self, *ss: str) -> "KObj":
        raise self._te()

    # deserialise content from stream
    def frombytes(self, data: bytes, offset: int) -> Tuple["KObj", int]:
        raise self._te()

    def __eq__(self, other: Any) -> bool:
        # this could be optimised in subclasses with appropriate descent logic
        if isinstance(other, KObj):
            return b9(self) == b9(other)
        return NotImplemented


# constructors always take type t, optional context, and
# optionally a size, attr pair
class KObjAtom(KObj):
    def __init__(
        self,
        t: int = 0,
        context: KContext = DEFAULT_CONTEXT,
        sz: int = 0,
        attr: int = 0,
    ) -> None:
        if t > 0 and t != TypeEnum.NIL:
            raise ValueError(f"Not atomic type {t}")
        super().__init__(t, context)
        self.data: bytes = b"\x00" * ATOM_LENGTH[-self.t]

    # atom setters
    def b(self, b: bool) -> KObj:
        if self.t not in [-TypeEnum.KB]:
            raise ValueError(f"wrong type {self._tn()} for g()")
        self.data = struct.pack("B", {True: 1, False: 0}[b])
        return self

    def c(self, c: str) -> KObj:
        bs = c.encode("ascii")
        if self.t not in [-TypeEnum.KC]:
            raise ValueError(f"wrong type {self._tn()} for c()")
        if len(bs) != 1:
            raise ValueError(".c() takes single character")
        self.data = bs
        return self

    def g(self, g: int) -> KObj:
        if self.t not in [-TypeEnum.KG]:
            raise ValueError(f"wrong type {self._tn()} for g()")
        self.data = struct.pack("B", g)
        return self

    def h(self, h: int) -> KObj:
        if self.t not in [-TypeEnum.KH]:
            raise ValueError(f"wrong type {self._tn()} for h()")
        self.data = struct.pack("h", h)
        return self

    def i(self, i: int) -> KObj:
        if self.t not in [-TypeEnum.KI]:
            raise ValueError(f"wrong type {self._tn()} for i()")
        self.data = struct.pack("i", i)
        return self

    def j(self, j: int) -> KObj:
        if self.t not in [-TypeEnum.KJ, -TypeEnum.KP]:
            raise ValueError(f"wrong type {self._tn()} for j()")
        self.data = struct.pack("q", j)
        return self

    def f(self, f: float) -> KObj:
        if self.t not in [-TypeEnum.KF, -TypeEnum.KZ]:
            raise ValueError(f"wrong type {self._tn()} for f()")
        self.data = struct.pack("d", f)
        return self

    def uu(self, uu: uuid.UUID) -> KObj:
        if self.t not in [-TypeEnum.UU]:
            raise ValueError(f"wrong type {self._tn()} for uu()")
        self.data = uu.bytes
        return self

    # atom getters
    def aB(self) -> bool:
        if self.t not in [-TypeEnum.KB]:
            raise ValueError(f"wrong type {self._tn()} for aB")
        return self.aG() == 1

    def aG(self) -> int:
        if self.t not in [-TypeEnum.KG, -TypeEnum.KB]:
            raise ValueError(f"wrong type {self._tn()} for aG")
        return cast(int, struct.unpack("B", self.data)[0])

    def aH(self) -> int:
        if self.t not in [-TypeEnum.KH]:
            raise ValueError(f"wrong type {self._tn()} for aH")
        return cast(int, struct.unpack("h", self.data)[0])

    def aI(self) -> int:
        if self.t not in [
            -TypeEnum.KI,
            -TypeEnum.KM,
            -TypeEnum.KD,
            -TypeEnum.KU,
            -TypeEnum.KV,
            -TypeEnum.KT,
        ]:
            raise ValueError(f"wrong type {self._tn()} for aI")
        return cast(int, struct.unpack("i", self.data)[0])

    def aJ(self) -> int:
        if self.t not in [-TypeEnum.KJ, -TypeEnum.KP, -TypeEnum.KN]:
            raise ValueError(f"wrong type {self._tn()} for aJ")
        return cast(int, struct.unpack("q", self.data)[0])

    def aU(self) -> uuid.UUID:
        if self.t not in [-TypeEnum.UU]:
            raise ValueError(f"wrong type {self._tn()} for aU")
        return uuid.UUID(bytes=self.data)

    def aE(self) -> float:
        if self.t not in [-TypeEnum.KE]:
            raise ValueError(f"wrong type {self._tn()} for aE")
        return cast(float, struct.unpack("f", self.data)[0])

    def aF(self) -> float:
        if self.t not in [-TypeEnum.KF, -TypeEnum.KZ]:
            raise ValueError(f"wrong type {self._tn()} for aF")
        return cast(float, struct.unpack("d", self.data)[0])

    def aC(self) -> str:
        if self.t not in [-TypeEnum.KC]:
            raise ValueError(f"wrong type {self._tn()} for aC")
        bs = struct.unpack("c", self.data)[0]
        return cast(str, bs.decode("ascii"))

    # serialisation
    def _databytes(self) -> bytes:
        return super()._databytes() + self.data

    def _paysz(self) -> int:
        assert len(self.data) == ATOM_LENGTH[-self.t]
        return super()._paysz() + ATOM_LENGTH[-self.t]

    def frombytes(self, data: bytes, offset: int) -> Tuple[KObj, int]:
        bs = ATOM_LENGTH[-self.t]
        self.data = data[offset : offset + bs]
        return self, offset + bs


class KSymAtom(KObj):
    def __init__(
        self,
        t: int = 0,
        context: KContext = DEFAULT_CONTEXT,
        sz: int = 0,
        attr: int = 0,
    ) -> None:
        super().__init__(t, context)
        self.data: bytes = b""

    def aI(self) -> int:
        return cast(int, struct.unpack("i", self.data)[0])

    def aS(self) -> str:
        return self.context.symbols_enc[self.aI()][0]

    def ss(self, s: str) -> KObj:
        if self.t not in [-TypeEnum.KS, TypeEnum.KRR]:
            raise ValueError(f"wrong type {self._tn()} for ss()")
        self.data = struct.pack("i", self.context.ss(s))
        return self

    # serialisation
    def _databytes(self) -> bytes:
        return super()._databytes() + self.context.symbols_enc[self.aI()][1]

    def _paysz(self) -> int:
        return super()._paysz() + len(self.context.symbols_enc[self.aI()][1])

    def frombytes(self, data: bytes, offset: int) -> Tuple[KObj, int]:
        bs = data[offset:].index(b"\x00") + 1
        self.ss(data[offset : offset + bs - 1].decode("ascii"))
        return self, offset + bs


class KErrAtom(KObj):
    def __init__(
        self,
        context: KContext = DEFAULT_CONTEXT,
    ) -> None:
        super().__init__(TypeEnum.KRR, context)
        self.data: bytes = b""

    def aS(self) -> str:
        return self.data[:-1].decode("ascii")

    def ss(self, s: str) -> KObj:
        self.data = bytes(s, "ascii") + b"\x00"
        return self

    # serialisation
    def _databytes(self) -> bytes:
        return super()._databytes() + self.data

    def _paysz(self) -> int:
        return super()._paysz() + len(self.data)

    def frombytes(self, data: bytes, offset: int) -> Tuple[KObj, int]:
        bs = data[offset:].index(b"\x00") + 1
        self.data = data[offset : offset + bs - 1]
        return self, offset + bs


class KRangedType(KObj):
    def frombytes(self, data: bytes, offset: int) -> Tuple[KObj, int]:
        attrib, sz = struct.unpack_from("<BI", data, offset=offset)
        self.attrib = attrib
        logger.debug(f" frombytes for ranged type attrib={attrib} sz={sz}")
        return self._ranged_frombytes(sz, data, offset + 5)

    def _ranged_frombytes(self, sz: int, data: bytes, offset: int) -> Tuple[KObj, int]:
        raise self._te()


class KByteArray(KRangedType):
    def __init__(self, t: int = TypeEnum.KG, sz: int = 0, attr: int = 0) -> None:
        super().__init__(t, attr=attr)
        self._g = array.array("B", [0] * sz)

    def _paysz(self) -> int:
        return 2 + 4 + 1 * len(self._g)

    def _databytes(self) -> bytes:
        return struct.pack("<bBI", self.t, self.attrib, len(self._g)) + struct.pack(
            f"<{len(self._g)}B", *self._g
        )

    def kG(self) -> "MutableSequence[int]":
        return self._g

    def kB(self) -> "MutableSequence[bool]":
        return BoolByteAdaptor(self._g)

    def __len__(self) -> int:
        return len(self._g)

    def _ranged_frombytes(self, sz: int, data: bytes, offset: int) -> Tuple[KObj, int]:
        self._g = array.array("B", data[offset : offset + sz])
        return self, offset + sz


class KShortArray(KRangedType):
    def __init__(self, t: int = TypeEnum.KH, sz: int = 0, attr: int = 0) -> None:
        super().__init__(t, attr=attr)
        self._h = array.array("h", [0] * sz)

    def _paysz(self) -> int:
        return 2 + 4 + 2 * len(self._h)

    def _databytes(self) -> bytes:
        return struct.pack("<bBI", self.t, self.attrib, len(self._h)) + struct.pack(
            f"<{len(self._h)}H", *self._h
        )

    def kH(self) -> "MutableSequence[int]":
        return self._h

    def __len__(self) -> int:
        return len(self._h)

    def _ranged_frombytes(self, sz: int, data: bytes, offset: int) -> Tuple[KObj, int]:
        self._h = array.array("h", data[offset : offset + 2 * sz])
        return self, offset + 2 * sz


class KIntArray(KRangedType):
    # python doesn't have a reliable 32-bit array type, "l" can either be 32 or 64 bits
    # which breaks direct cast-style deserialising. So when l is 64 bits we need to
    # iteratively unpack with unpack_from. Somewhat annoying given the other types map
    # cleanly.
    def __init__(self, t: int = TypeEnum.KI, sz: int = 0, attr: int = 0) -> None:
        super().__init__(t, attr=attr)
        self._i = array.array("l", [0] * sz)

    def _paysz(self) -> int:
        return 2 + 4 + 4 * len(self._i)

    def _databytes(self) -> bytes:
        if self._i.itemsize == 4:  # array type 'l' seems to store as 8 bytes
            pi = self._i.tobytes()
        else:
            pi = struct.pack(f"<{len(self._i)}I", *self._i)
        return struct.pack("<bBI", self.t, self.attrib, len(self._i)) + pi

    def kI(self) -> "MutableSequence[int]":
        return self._i

    def __len__(self) -> int:
        return len(self._i)

    def _ranged_frombytes(self, sz: int, data: bytes, offset: int) -> Tuple[KObj, int]:
        o2 = offset + 4 * sz
        if self._i.itemsize == 4:
            self._i = array.array("l", data[offset:o2])
        else:
            self._i = array.array("l", [0] * sz)
            for i in range(sz):
                self._i[i] = struct.unpack_from("<l", data, offset + 4 * i)[0]
        return self, o2


class KIntSymArray(KIntArray):
    # store symbol indexes in KIntArray
    # hook serialise to use null byte terminated representation
    def _paysz(self) -> int:
        return 2 + 4 + sum([len(self.context.symbols_enc[j][1]) for j in self._i])

    def _databytes(self) -> bytes:
        parts = [struct.pack("<bBI", self.t, self.attrib, len(self._i))]
        for j in self._i:
            parts.append(self.context.symbols_enc[j][1])
        return b"".join(parts)

    def kS(self) -> "Sequence[str]":
        # Warning: accessor read-only
        s = []
        for j in self._i:
            s.append(self.context.symbols_enc[j][0])
        return s

    def appendS(self, *ss: str) -> KObj:
        for s in ss:
            j = self.context.ss(s)
            self._i.append(j)
        return self

    def __len__(self) -> int:
        return len(self._i)

    def _ranged_frombytes(self, sz: int, data: bytes, offset: int) -> Tuple[KObj, int]:
        self._i = array.array("l", [0] * sz)
        for i in range(sz):
            bs = data[offset:].index(b"\x00") + 1
            s = data[offset : offset + bs - 1].decode("ascii")
            d = self.context.ss(s)
            self._i[i] = d
            offset += bs
        return self, offset


class KLongArray(KRangedType):
    def __init__(self, t: int = TypeEnum.KJ, sz: int = 0, attr: int = 0) -> None:
        super().__init__(t, attr=attr)
        self._j = array.array("q", [0] * sz)

    def _paysz(self) -> int:
        return 2 + 4 + 8 * len(self._j)

    def _databytes(self) -> bytes:
        return struct.pack("<bBI", self.t, self.attrib, len(self._j)) + struct.pack(
            f"<{len(self._j)}q", *self._j
        )

    def kJ(self) -> "MutableSequence[int]":
        return self._j

    def __len__(self) -> int:
        return len(self._j)

    def _ranged_frombytes(self, sz: int, data: bytes, offset: int) -> Tuple[KObj, int]:
        self._j = array.array("q", data[offset : offset + 8 * sz])
        return self, offset + 8 * sz


class KFloatArray(KRangedType):
    def __init__(self, t: int = TypeEnum.KF, sz: int = 0, attr: int = 0) -> None:
        super().__init__(t, attr=attr)
        self._e = array.array("f", [0] * sz)

    def _paysz(self) -> int:
        return 2 + 4 + 4 * len(self._e)

    def _databytes(self) -> bytes:
        return struct.pack("<bBI", self.t, self.attrib, len(self._e)) + struct.pack(
            f"<{len(self._e)}f", *self._e
        )

    def kE(self) -> "MutableSequence[float]":
        return self._e

    def __len__(self) -> int:
        return len(self._e)

    def _ranged_frombytes(self, sz: int, data: bytes, offset: int) -> Tuple[KObj, int]:
        self._e = array.array("f", data[offset : offset + 4 * sz])
        return self, offset + 4 * sz


class KDoubleArray(KRangedType):
    def __init__(self, t: int = TypeEnum.KF, sz: int = 0, attr: int = 0) -> None:
        super().__init__(t, attr=attr)
        self._f = array.array("d", [0] * sz)

    def _paysz(self) -> int:
        return 2 + 4 + 8 * len(self._f)

    def _databytes(self) -> bytes:
        return struct.pack("<bBI", self.t, self.attrib, len(self._f)) + struct.pack(
            f"<{len(self._f)}d", *self._f
        )

    def kF(self) -> "MutableSequence[float]":
        return self._f

    def __len__(self) -> int:
        return len(self._f)

    def _ranged_frombytes(self, sz: int, data: bytes, offset: int) -> Tuple[KObj, int]:
        self._f = array.array("d", data[offset : offset + 8 * sz])
        return self, offset + 8 * sz


class KCharArray(KRangedType):
    def __init__(self, t: int = TypeEnum.KC, sz: int = 0, attr: int = 0) -> None:
        super().__init__(t, attr=attr)
        self._c: array.array[str] = array.array("u", [" "] * sz)

    def _paysz(self) -> int:
        return 2 + 4 + 1 * len(self._c)

    def _databytes(self) -> bytes:
        bs = self._c.tounicode().encode("ascii")
        return struct.pack("<bBI", self.t, self.attrib, len(self._c)) + bs

    def kC(self) -> array.array:  # type: ignore[type-arg]
        return self._c

    def aS(self) -> str:
        return self._c.tounicode()

    def __len__(self) -> int:
        return len(self._c)

    def _ranged_frombytes(self, sz: int, data: bytes, offset: int) -> Tuple[KObj, int]:
        s = data[offset : offset + sz].decode("ascii")
        self._c = array.array("u", [])
        self._c.fromunicode(s)
        return self, offset + sz


class KObjArray(KRangedType):
    def __init__(self, t: int = TypeEnum.K) -> None:
        super().__init__(0)
        self._k: List[KObj] = []

    def _paysz(self) -> int:
        # sum sizes nested ks
        return 2 + 4 + sum([ko._paysz() for ko in self._k])

    def _databytes(self) -> bytes:
        parts = [struct.pack("<bBI", self.t, self.attrib, len(self._k))]
        parts.extend(ko._databytes() for ko in self._k)
        return b"".join(parts)

    def kK(self) -> "MutableSequence[KObj]":
        return self._k

    def __len__(self) -> int:
        return len(self._k)

    def _ranged_frombytes(self, sz: int, data: bytes, offset: int) -> Tuple[KObj, int]:
        for i in range(sz):
            obj, offset = _d9_unpackfrom(data, offset)
            self._k.append(obj)
        return self, offset


class KUUIDArray(KRangedType):
    def __init__(self, t: int = TypeEnum.UU, sz: int = 0, attr: int = 0) -> None:
        super().__init__(TypeEnum.UU)
        self._u: List[uuid.UUID] = [uuid.UUID(int=0)] * sz

    def _paysz(self) -> int:
        # sum sizes nested ks
        return 2 + 4 + 16 * len(self._u)

    def _databytes(self) -> bytes:
        parts = [struct.pack("<bBI", self.t, self.attrib, len(self._u))]
        parts.extend(uu.bytes for uu in self._u)
        return b"".join(parts)

    def kU(self) -> "MutableSequence[uuid.UUID]":
        return self._u

    def __len__(self) -> int:
        return len(self._u)

    def _ranged_frombytes(self, sz: int, data: bytes, offset: int) -> Tuple[KObj, int]:
        for i in range(sz):
            start = offset + i * 16
            self._u.append(uuid.UUID(bytes=data[start : start + 16]))
        return self, offset + 16 * sz


def b9(k: KObj, msgtype: int = 0, flags: int = 0) -> bytes:
    # 8 byte header
    msglen = 8 + k._paysz()
    return struct.pack("<BBHI", 1, msgtype, flags, msglen) + k._databytes()


def d9(data: bytes) -> KObj:
    # raises struct.error on underflow
    ver, msgtype, flags, msglen = struct.unpack_from("<BBHI", data, offset=0)
    if len(data) < msglen:
        raise ValueError(
            f"buffer is too short, required {msglen} bytes, got {len(data)}"
        )
    offset = 8
    if flags == 1:
        print(data.hex())
        data = decompress(data[8:])
        offset = 0
        msglen = len(data)
    elif flags != 0:
        print(data.hex())
        raise ValueError(
            f"unknown payload flags={flags} - not yet implemented, please open an Issue"
        )

    k, pos = _d9_unpackfrom(data, offset=offset)
    if pos != msglen:
        raise Exception(f"Final position at {pos} expected {msglen}")
    return k


def _d9_unpackfrom(data: bytes, offset: int) -> Tuple[KObj, int]:
    (t,) = struct.unpack_from("<b", data, offset=offset)
    offset += 1
    logger.debug(f" at offset {offset}/{len(data)} unpacking type {tn(t)}")
    if t == -TypeEnum.KS or t == TypeEnum.KRR:
        return KSymAtom(t).frombytes(data, offset)
    elif t < 0:
        return KObjAtom(t).frombytes(data, offset)
    elif t >= 0 and t < 20:
        # ranged vector types, need to verify t
        return VECTOR_CONSTUCTORS[t](t).frombytes(data, offset)
    elif t == TypeEnum.NIL:
        # seems to be a null byte after the type?
        return KObjAtom(t).frombytes(data, offset)
    elif t == TypeEnum.XD:
        kkeys, offset = _d9_unpackfrom(data, offset)
        kvalues, offset = _d9_unpackfrom(data, offset)
        return KDict(kkeys, kvalues), offset
    elif t == TypeEnum.XT:
        kkeys, offset = _d9_unpackfrom(data, offset + 1)
        return KFlip(kkeys), offset
    elif t == TypeEnum.SD:
        kkeys, offset = _d9_unpackfrom(data, offset)
        kvalues, offset = _d9_unpackfrom(data, offset)
        return KDict(kkeys, kvalues, t), offset
    elif t >= 20 and t < 30:
        return VECTOR_CONSTUCTORS[TypeEnum.KJ](t).frombytes(data, offset)
    raise ValueError(f"Unable to d9 unpack t={t}")


# atom constructors
def ka(t: Union[int, TypeEnum]) -> KObj:
    return KObjAtom(t)


def kb(b: bool) -> KObj:
    return KObjAtom(-TypeEnum.KB).b(b)


def kc(c: str) -> KObj:
    return KObjAtom(-TypeEnum.KC).c(c)


def kg(i: int) -> KObj:
    return KObjAtom(-TypeEnum.KG).g(i)


def kh(i: int) -> KObj:
    return KObjAtom(-TypeEnum.KH).h(i)


def ki(i: int) -> KObj:
    return KObjAtom(-TypeEnum.KI).i(i)


def kf(f: float) -> KObj:
    return KObjAtom(-TypeEnum.KF).f(f)


def kj(i: int) -> KObj:
    return KObjAtom(-TypeEnum.KJ).j(i)


def kp(i: int) -> KObj:
    return KObjAtom(-TypeEnum.KP).j(i)


def ks(s: str) -> KObj:
    return KSymAtom(-TypeEnum.KS).ss(s)


def kuu(uu: uuid.UUID) -> KObj:
    return KObjAtom(-TypeEnum.UU).uu(uu)


# vector constructors
VECTOR_CONSTUCTORS: Dict[TypeEnum, Type[KObj]] = {
    TypeEnum.K: KObjArray,
    TypeEnum.KB: KByteArray,
    TypeEnum.KG: KByteArray,
    TypeEnum.KH: KShortArray,
    TypeEnum.KI: KIntArray,
    TypeEnum.KJ: KLongArray,
    TypeEnum.KS: KIntSymArray,
    TypeEnum.UU: KUUIDArray,
    TypeEnum.KC: KCharArray,
    TypeEnum.KP: KLongArray,
    TypeEnum.KF: KDoubleArray,
    TypeEnum.KE: KFloatArray,
    TypeEnum.KM: KIntArray,
    TypeEnum.KD: KIntArray,
    TypeEnum.KZ: KDoubleArray,
    TypeEnum.KN: KLongArray,
    TypeEnum.KU: KIntArray,
    TypeEnum.KV: KIntArray,
    TypeEnum.KT: KIntArray,
}


def ktn(t: TypeEnum, sz: int = 0, attr: AttrEnum = AttrEnum.NONE) -> KObj:
    if t == TypeEnum.K:
        if sz > 0:
            raise ValueError("ktn K can only be empty at initialisation")
        return KObjArray(t)
    try:
        return VECTOR_CONSTUCTORS[t](t, sz=sz, attr=attr)
    except KeyError:
        raise ValueError(f"ktn for type {tn(t)}")


def kk(*objs: KObj) -> KObj:
    for o in objs:
        if not isinstance(o, KObj):
            raise ValueError(f"not KObj: {o}")
    k = ktn(TypeEnum.K)
    k.kK().extend(objs)
    return k


def cv(s: str) -> KObj:
    k = ktn(TypeEnum.KC)
    k.kC().fromunicode(s)
    return k


class KDict(KObj):
    def __init__(self, kkeys: KObj, kvalues: KObj, t: TypeEnum = TypeEnum.XD):
        if len(kkeys) != len(kvalues):
            raise ValueError("dict keys and values must be same length")
        super().__init__(t)
        if t == TypeEnum.SD and kkeys.t != TypeEnum.XT and kkeys.attrib == 0:
            raise ValueError(f"Keys not sorted for SD {kkeys._tn()}")
        self._kkey = kkeys
        self._kvalue = kvalues

    def _paysz(self) -> int:
        return 1 + self._kkey._paysz() + self._kvalue._paysz()

    def _databytes(self) -> bytes:
        return (
            struct.pack("<B", self.t)
            + self._kkey._databytes()
            + self._kvalue._databytes()
        )

    def __len__(self) -> int:
        return len(self._kkey)

    def kkey(self) -> KObj:
        return self._kkey

    def kvalue(self) -> KObj:
        return self._kvalue

    def kS(self) -> "Sequence[str]":
        # key names
        if self._kkey.t == TypeEnum.KS:
            return self.kkey().kS()
        elif self._kkey.t == TypeEnum.K:
            return [k.aS() for k in self.kkey().kK()]

        raise KeyError(f"Keys as strings not possible on {self._kkey._tn()}")

    def __getitem__(self, item: Union[int, str]) -> KObj:
        if not isinstance(item, str):
            raise ValueError("cannot index dict by int")
        if self._kkey.t == TypeEnum.KS:
            try:
                idx = self.kkey().kS().index(item)
                return self.kvalue().kK()[idx]
            except ValueError:
                raise KeyError(f"Key not found {item}")
        elif self._kkey.t == TypeEnum.K:
            # check for any nested atomic symbol keys within the K object array
            for idx in range(len(self._kkey)):
                k = self._kkey.kK()[idx]
                if k.t == -TypeEnum.KS:
                    if k.aS() == item:
                        return self.kvalue().kK()[idx]
            raise KeyError(f"Key not found {item}")

        raise KeyError(f"Key lookup not possible on {self._kkey._tn()}")


class KFlip(KObj):
    def __init__(self, kd: KObj, sorted: bool = False):
        if kd.t != TypeEnum.XD:
            raise ValueError(f"can only flip a dict, not {kd._tn()}")
        if len(kd.kkey()) == 0:
            raise ValueError("must have >0 columns")
        if kd.kvalue().t != TypeEnum.K:
            raise ValueError("must have K vector holding cols")
        if kd.kkey().t != TypeEnum.KS:
            raise ValueError("dict key must be S vector of column names")
        if len(kd.kvalue().kK()) == 0:
            raise ValueError("cannot have zero columns")
        # check all columns have same length
        c = len(kd.kvalue().kK()[0])
        for col in kd.kvalue().kK():
            if len(col) != c:
                raise ValueError("column length inconsistent")

        attr = AttrEnum.NONE
        if sorted:
            attr = AttrEnum.SORTED
        super().__init__(TypeEnum.XT, attr=attr)
        self._kvalue = kd

    def _paysz(self) -> int:
        return 2 + self._kvalue._paysz()

    def _databytes(self) -> bytes:
        return struct.pack("<bB", self.t, self.attrib) + self._kvalue._databytes()

    def kvalue(self) -> KObj:
        return self._kvalue

    def __len__(self) -> int:
        # the length of a table is not the length of it's contained dictionary, as that is the number of columns.
        # we need to read-through to the first columns data.
        assert self.kvalue().t == TypeEnum.XD
        tdict_values = self.kvalue().kvalue()
        if len(tdict_values) == 0:
            raise ValueError("flip contains a dict with no values, so has no length")
        assert tdict_values.t == TypeEnum.K
        first_col = tdict_values.kK()[0]
        return len(first_col)

    def __getitem__(self, item: Union[int, str]) -> KObj:
        if isinstance(item, int):
            if item >= len(self):
                raise IndexError()
            d = self.kvalue()
            vals = []
            for k in self.kS():
                v = self[k]
                # tricky: get index item from v
                if v.t == TypeEnum.KJ:
                    vals.append(kj(v.kJ()[item]))
                elif v.t == TypeEnum.KI:
                    vals.append(ki(v.kI()[item]))
                elif v.t == TypeEnum.KH:
                    vals.append(kh(v.kH()[item]))
                elif v.t == TypeEnum.KB:
                    vals.append(kb(v.kB()[item]))
                elif v.t == TypeEnum.UU:
                    vals.append(kuu(v.kU()[item]))
                elif v.t == TypeEnum.KG:
                    vals.append(kg(v.kG()[item]))
                elif v.t == TypeEnum.KF:
                    vals.append(kf(v.kF()[item]))
                elif v.t == TypeEnum.KG:
                    vals.append(kg(v.kG()[item]))
                elif v.t == TypeEnum.KC:
                    vals.append(kc(v.kC()[item]))
                elif v.t == TypeEnum.KS:
                    vals.append(ks(v.kS()[item]))
                elif v.t == TypeEnum.KP:
                    vals.append(kp(v.kJ()[item]))
                elif v.t == TypeEnum.K:
                    vals.append(v.kK()[item])
                else:
                    raise ValueError(f"no boxing defined for {v}")

            return xd(d.kkey(), kk(*vals))
        elif isinstance(item, str):
            try:
                idx = self.kS().index(item)
                return self.kK()[idx]
            except ValueError:
                raise KeyError(f"Column not found {item}")

    def kS(self) -> "Sequence[str]":
        # column names
        return self.kvalue().kkey().kS()

    def kK(self) -> "MutableSequence[KObj]":
        # column values
        return self.kvalue().kvalue().kK()


def krr(msg: str) -> KObj:
    return KSymAtom(TypeEnum.KRR).ss(msg)


kNil = ka(TypeEnum.NIL)


class KException(Exception):
    pass


def xd(kkeys: KObj, kvalues: KObj, sorted: bool = False) -> KDict:
    if sorted:
        return KDict(kkeys, kvalues, TypeEnum.SD)
    return KDict(kkeys, kvalues)


def xt(kd: KDict, sorted: bool = False) -> KFlip:
    return KFlip(kd, sorted=sorted)



import pytest

from aiokdb import KContext


def test_context() -> None:
    kcon = KContext()
    assert kcon.ss("hello") == 0
    assert kcon.ss("world") == 1
    assert kcon.ss("hello") == 0
    assert kcon.ss("world") == 1
    assert kcon.ss("how") == 2
    with pytest.raises(ValueError):
        kcon.ss("💩")
    assert len(kcon.symbols) == 3



import argparse
import logging
import random
import struct
import sys
from typing import Optional

logger = logging.getLogger(__name__)


def decompress(data: bytes) -> bytes:
    # decompression based on c.java
    # https://github.com/KxSystems/javakdb/blob/master/javakdb/src/main/java/com/kx/c.java#L667
    # which in turn seems quite similar to LZRW1 http://www.ross.net/compression/lzrw1.html
    # however with a much smaller window for the hash function (2 bytes) and only 256 slots

    uncomp_sz = struct.unpack("I", data[0:4])[0] - 8
    logger.debug(f"decompressing: {len(data)} compressed bytes to {uncomp_sz} bytes")

    # compressed stream repeats: control byte (8 single-bit instructions), followed by 8-16 data bytes.
    # Bits of the control byte are examined bit-by-bit from least significant:
    #   control bit clear: copy one literal byte to output.
    #   control bit set: read 2 bytes [index, sz]. Copy from hashed history table entry [index] for sz bytes.
    # After each operation, update the hashpos table provided we have at least 2 bytes output, and using no more
    # than two bytes from the history copy

    dst = bytearray(uncomp_sz)
    hashpos = [0] * 256
    d = 4  # position index into compressed data[]
    s = 0  # write position in dst[] (uncompressed output)
    p = 0  # hash  position in dst[] for hashpos updates
    f = 0  # current instruction byte, examined bit by bit
    i = 0  # bit position within f as a 1-hot value, ie. 1,2,4,8,16,32,64,128

    while s < uncomp_sz:
        if i == 0:
            # out of instruction bits! reload f, reset i, advance d
            # print(f"loading next 8 instructions from d={d}")
            f = data[d]
            d += 1
            i = 1

        # check bit i of f
        if f & i:  # bit set: copy history range
            # {ptr}{sz} copy n bytes (sz+2) from uncompressed history pointer r (ptr)
            r = hashpos[data[d]]
            n = data[d + 1]
            # print(f"instr {i} hist d={d} slot {data[d]}, copy pos={r} len={2+n} to {s}")
            # DO NOT USE SLICE ASSIGNMENT HERE, AS A HISTORY REFERENCE CAN COPY OWN OUTPUT
            # ie. start 8 bytes back and copy 64 bytes, gives 8x repeating 8 bytes
            for m in range(2 + n):
                dst[s + m] = dst[r + m]
            d += 2
            s += 2

        else:  # copy 1 byte from compressed stream to uncomp
            # print(f"instr {i} literal d={d} value {data[d:d+1].hex()} to pos {s}")
            dst[s] = data[d]
            n = 0
            d += 1
            s += 1

        while p < s - 1:
            hashv = dst[p] ^ dst[p + 1]
            # print(f" hash slot {hashv} set to p={p}")
            hashpos[hashv] = p
            p += 1

        # only first two bytes of a copied range are used to update the hash
        # jump s and p over the remaining copied range (n)
        s += n
        if f & i:
            p = s

        # next control bit
        i = (i << 1) & 255

    # Note that cpython is overly restrictive in the uuid.UUID(bytes=...) call and will not
    # allow a bytearray, so now decompression is done convert it once to immutable bytes object.
    #  https://github.com/python/cpython/blob/8edfa0b0b4ae4235bb3262d952c23e7581516d4f/Lib/uuid.py#L188
    return bytes(dst)


def compress(y: bytes) -> Optional[bytes]:
    # if data is not worth compressing return None, else compressed payload
    if len(y) < 8:
        return None

    wr = bytearray(len(y) // 2)

    i = 128  # bit position within f as a 1-hot value, ie. 1,2,4,8,16,32,64,128
    f = 0  # current flags
    h0 = 0
    h = 0
    p = 0
    s0 = 0
    s = 0  # read position in uncompressed y
    t = len(y)
    a = [0] * 256  # hash buckets

    struct.pack_into("I", wr, 0, len(y) + 8)
    d = 4  # write position into wr
    c = 4  # last flags position into wr (sentinal first loop, written to twice)

    while s < t:
        i *= 2
        if i == 256:
            # print(f"flushing flags wBuf[{c}]={f} writepos={d}")
            if d > len(wr) - 17:  # 8x2 data + 1 flags
                logger.debug(
                    f"compressor: {len(y)} bytes not worth compressing (writepos {d})"
                )
                return None
            i = 1
            wr[c] = f  # write back flags
            c = d
            d += 1
            f = 0

        # h0, s0 are hash and position writeback?
        # h current hash
        if s > t - 3:
            literal = True
        else:
            h = y[s] ^ y[s + 1]
            p = a[h]
            # quirk: it doesn't check y[s+1]==y[p+1]. Since h is over two bytes,
            # if hash matches and y[s]==y[p], it cannot be a collision
            literal = (0 == p) or (y[s] != y[p])

        # print(f"operation literal={literal} writepos={d} readpos={s}")
        if 0 < s0:
            a[h0] = s0
            s0 = 0

        if literal:
            # write back the (possibly updated) backref for this hash
            h0 = h
            s0 = s
            # cp the literal. f flag already clear
            wr[d] = y[s]
            d += 1
            s += 1
        else:
            a[h] = s
            f |= i
            p += 2
            s += 2
            # test match length
            r = s
            q = min(s + 255, t) - 1
            # print(f"flags now {f} maxlen {q}")
            while y[p] == y[s] and s < q:
                p += 1
                s += 1
            # print(f'  runlength {s-r}')
            wr[d + 0] = h
            wr[d + 1] = s - r
            d += 2

    wr[c] = f  # flush flags
    logger.debug(f"compressor: {len(y)} uncompressed bytes down to {d} bytes")
    return wr[0:d]


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--debug", action="store_true")

    args = parser.parse_args()
    debug_level = {True: logging.DEBUG, False: logging.INFO}
    logging.basicConfig(level=debug_level[args.debug])

    while True:
        sz = random.randrange(1, 10000)
        spots = random.randrange(0, sz)  # patch in some non-zeros
        data = bytearray(sz)
        for i in range(spots):
            data[random.randrange(1, sz)] = random.randrange(0, 255)

        bs = compress(data)
        if bs is not None:
            ds = decompress(bs)
            if not ds == data:
                print("mismatch")
                print(data)
                print(bs)
                print(ds)
                sys.exit(-1)



from collections.abc import MutableSequence, Sequence
from typing import TYPE_CHECKING, Any, Iterable, Union, overload

if TYPE_CHECKING:
    BaseMutSeq = MutableSequence[bool]
else:
    BaseMutSeq = MutableSequence


class BoolByteAdaptor(BaseMutSeq):
    def __init__(self, data: "MutableSequence[int]"):
        self.data = data

    @overload
    def __getitem__(self, i: int) -> bool: ...
    @overload
    def __getitem__(self, s: slice) -> "MutableSequence[bool]": ...
    def __getitem__(self, i: Union[int, slice]) -> Union[bool, "MutableSequence[bool]"]:
        if isinstance(i, slice):
            return self.__class__(self.data[i])
        else:
            return {0: False, 1: True}[self.data[i]]

    def __len__(self) -> int:
        return len(self.data)

    @overload
    def __setitem__(self, index: int, item: bool) -> None: ...
    @overload
    def __setitem__(self, index: slice, item: Iterable[bool]) -> None: ...
    def __setitem__(
        self, index: Union[int, slice], item: Union[bool, Iterable[bool]]
    ) -> None:
        if isinstance(index, slice) and isinstance(item, Iterable):
            self.data[index] = [{True: 1, False: 0}[i] for i in item]
        elif isinstance(index, int) and isinstance(item, bool):
            self.data[index] = {True: 1, False: 0}[item]
        else:
            raise TypeError()

    def insert(self, index: int, item: bool) -> None:
        self.data.insert(index, {True: 1, False: 0}[item])

    def __delitem__(self, item: Union[int, slice]) -> None:
        del self.data[item]

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Sequence):
            if len(self) == len(other):
                return all([a == b for a, b in zip(self, other)])
            else:
                return False
        raise TypeError()
